/*
 * Copyright (C) 1995  Sam Moore, Warren Jasper
 * All rights reserved.
 *
 * This software may be freely copied, modified, and redistributed
 * provided that this copyright notice is preserved on all copies.
 *
 * You may not distribute this software, in whole or in part, as part of
 * any commercial product without the express consent of the authors.
 *
 * There is no warranty or other guarantee of fitness of this software
 * for any purpose.  It is provided solely "as is".
 *
 */

/***************************************************************************
 *
 * a2dc.c
 *
 ***************************************************************************/

#include <linux/autoconf.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/major.h>
#include <linux/mm.h>
#include <linux/timer.h>
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/signal.h>
#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <stdlib.h>
#include "adc.h"
#include "a2dc.h"

/***************************************************************************
 *
 * Prototype of pubic and private functions.
 *
 ***************************************************************************/

static int  SoftTrigRead(WORD *outBuff, ChanRec *chan);
static int  PacerRead(WORD *outBuff, ChanRec *chan);
static void SetGains(BYTE gain);
static int  SetPacerFreq(BoardRec *board);
static void StopPacer();
static int  SetChannelMux(BYTE lowChan, BYTE highChan);
static void LoadPacer(BoardRec *board);
static void LoadCounter0( LONG arg );

static int adc_read(struct inode *iNode, struct file *filePtr, char *buf, int count);
static int dio_write(struct inode *iNode, struct file *filePtr, char *buf, int count);
static int adc_open(struct inode *iNode, struct file *filePtr);
static void adc_close(struct inode *iNode, struct file *filePtr);
static int adc_ioctl(struct inode *iNode, struct file *filePtr, unsigned int cmd, LONG arg);
static void adc0_ReadInterrupt();
static void adc0_TimerHandler(LONG);

int init_module(void);
void cleanup_module(void) ;

/***************************************************************************
 *
 * Global data.
 *
 ***************************************************************************/

static int WordsToRead;                     /* number of conversions until done  */
static ChanRec *CurrChan = NULL;            /* pointer to Chan[minor]            */
static int MajorNumber = DEFAULT_MAJOR_DEV; /* Major number compiled in          */
static ChanRec Chan[AD_CHANNELS];           /* Channel specific information      */
static BoardRec BoardData;                  /* Board specific information        */
static WORD KernBuff[2048];                 /* Kernel buffer where samples are   */
                                            /* saved before write to user space  */
                                            /* structure for timer data          */
static WORD* KernBuffPtr;                   /* pointer to kernbuf                */

char kernel_version[] = UTS_RELEASE;        /* Required to by module             */
struct wait_queue  *adc0_wait;              /* wait semaphore                    */
struct timer_list TimerList = {NULL, NULL, 0, 0, adc0_TimerHandler};

/***************************************************************************
 *
 * read() service function
 *
 ***************************************************************************/

static int adc_read(struct inode *iNode, struct file *filePtr, char *buf, int count)
{
  int	  stat;
  BYTE  value;
  int   minor = MINOR(iNode->i_rdev);
  register int i;

  if ( minor >= 0 && minor < AD_CHANNELS ) {

    /* Set modes for this channel */

    SetGains((BYTE)Chan[minor].gain);
    Chan[minor].count = count;
    SetChannelMux(minor, minor); /* Read only this channel */

    /* Read */

    switch (Chan[minor].mode) {
      case ADC_SOFT_TRIGGER:
          #ifdef DEBUG
              printk("adcread(): Entering ADC_SOFT_TRIGGER mode.\n");
          #endif
          if (SoftTrigRead(KernBuff, &Chan[minor])) {
              printk("adc_read: SoftTrigRead() failed.\n");
              return(-1);
          } 
          break;
      case ADC_EXTERNAL_TRIGGER:
          #ifdef DEBUG
              printk("adcread(): Entering ADC_EXTERNAL_TRIGGER mode.\n");
          #endif
          if (PacerRead(KernBuff, &Chan[minor])) {
              printk("adc_read: ExternalRead() failed.\n");
              return(-1);
          }
          break;
      case ADC_PACER_CLOCK:
          #ifdef DEBUG
              printk("adcread(): Entering ADC_PACER_CLOCK mode.\n");
          #endif
          StopPacer();                           /* disable pacer if in pacer mode */
          outb_p(0x1, PACER_CLOCK_REG);	         /*IP0 controls counter gate*/
          if (PacerRead(KernBuff, &Chan[minor])) {
              printk("adc_read: PacerRead() failed.\n");
              return(-1);
          } 
          break;

      case ADC_PACER_EXTERN_TRIG:
          #ifdef DEBUG
              printk("adcread(): Entering ADC_PACER_EXTERN_TRIG mode.\n");
          #endif
          StopPacer();                           /* disable pacer if in pacer mode */
          outb_p(0x1, PACER_CLOCK_REG);	         /*IP0 controls counter gate*/
          if (PacerRead(KernBuff, &Chan[minor])) {
              printk("adc_read: PacerRead() with external tigger failed.\n");
              return(-1);
          } 
          break;
    }

    /* Check that data can be written to file. */

    if ((stat = verify_area(VERIFY_WRITE, buf, sizeof(WORD)*Chan[minor].count)) != 0) {
      printk("adc_read: Failed VERIFY_WRITE\n");
      return -1;
    }

    /* Write data to user space */

    for (i = 0; i < Chan[minor].count; i++) {
      KernBuff[i] >>= 4;
    }

    if (Chan[minor].count == 1) {
      cli();
      put_fs_word(*KernBuff, (WORD*) buf);
      sti();
    } else  {
      cli();
      memcpy_tofs(buf, KernBuff, Chan[minor].count*sizeof(WORD));
      sti();
    }

    return(Chan[minor].count);     /* return number of samples read */
  }

  /* check to see if reading a value from the DIO */
  if ( minor == AD_CHANNELS ) {
    value = inb(DIO_REG);
    cli();
    put_fs_byte(value, buf);
    sti();
    return 1;
  }

  printk("adc_read: Incorrect minor number (%d).\n", minor);
  return -1;
}

/***************************************************************************
 *
 * write() service function
 *
 ***************************************************************************/

static int dio_write(struct inode *iNode, struct file *filePtr, char *buf, int count)
{
    int  minor = MINOR(iNode->i_rdev);
    BYTE value;

    if ( minor != AD_CHANNELS ) {
        printk("dio_write: bad minor number = %d\n", minor );
        return -1;
    }
    cli();
    value = get_fs_byte(buf) & 0xf;
    outb_p( value, DIO_REG);
    sti();
    #ifdef DEBUG
        printk("DIO set to %#x\n", value & 0xf);
    #endif
    return 1;
}


/***************************************************************************
 *
 * open() service handler
 *
 ***************************************************************************/

static int adc_open(struct inode *iNode, struct file *filePtr)
{
    int minor = MINOR(iNode->i_rdev);

    MOD_INC_USE_COUNT;

    /* 
       check if device is already open: only one process may read from a
       port at a time.  There is still the possibility of two processes 
       reading from two different channels messing things up. However,
       the overhead to check for this may not be worth it.
    */

    if ( Chan[minor].open == TRUE ) {
        return -EBUSY;
    }

    Chan[minor].open = TRUE;              /* The device is open */
    Chan[minor].gain = BP_10_00V;         /* +/- 10V */
    Chan[minor].mode = filePtr->f_flags;  /* set trigger mode */

    if (Chan[minor].mode == ADC_PACER_EXTERN_TRIG) {
        BoardData.pacerReg = 0x1;
    } else {
        BoardData.pacerReg = 0x0;
    }

    outb_p(0x00, STATUS_REG);             /* Clear INT bit */
    outb_p(0x0, MISC_REG);                /* Put in compatibility mode */

    #ifdef DEBUG
        printk("%s: open(): minor %d mode %d.\n", ADAPTER_ID, minor, Chan[minor].mode);
    #endif
    return 0;   
}

/***************************************************************************
 *
 * close() service handler
 *
 ***************************************************************************/

static void adc_close(struct inode *iNode, struct file *filePtr)
{
    int minor = MINOR(iNode->i_rdev);

    #ifdef DEBUG
        printk("%s: close() of minor number %d.\n", ADAPTER_ID, minor);
    #endif

    MOD_DEC_USE_COUNT;
    Chan[minor].open = FALSE;
}

/***************************************************************************
 *
 * iotctl() service handler
 *
 ***************************************************************************/

/* Note:
    Remember that FIOCLEX, FIONCLEX, FIONBIO, and FIOASYN are reserved ioctl cmd numbers
*/

static int adc_ioctl(struct inode *iNode, struct file *filePtr, unsigned int cmd, LONG arg)
{
  int minor = MINOR(iNode->i_rdev);

  switch (cmd) {
    case ADC_SET_GAINS:
      Chan[minor].gain = (BYTE)arg;
      break;
    case ADC_SET_PACER_FREQ:
      if ( BoardData.freq > MAX_FREQ  && Chan[minor].mode == ADC_PACER_CLOCK) {
          printk("ioctl:  BoardData.freq = %ld out of range.\n", BoardData.freq);
          return -1;
      } else {
          BoardData.freq = (LONG)arg;
          SetPacerFreq(&BoardData);
          LoadPacer(&BoardData);               /* load the board frequency now */
          outb_p(BoardData.pacerReg, PACER_CLOCK_REG);  
      }
      break;
    case ADC_GET_PACER_FREQ:
      cli();
      put_fs_long(BoardData.freq, (long*) arg);
      sti();
      break;
    case ADC_STOP_PACER:
      StopPacer();
      break;
    case CTR0:
      if ( arg == 1 ) {
          BoardData.pacerReg |= 0x2;
      } else {
          BoardData.pacerReg &= ~(0x2);
      }
      outb_p(BoardData.pacerReg, PACER_CLOCK_REG);  
      break;
    case COUNTER0:
      LoadCounter0( (LONG) arg );
      break;
    case ADC_DIO_PRESET:
      if ( arg == -1 ) {
          Chan[minor].adc_dio_pretrig = FALSE;
      } else {
          Chan[minor].adc_dio_pretrig = TRUE;
          Chan[minor].dio_value = (BYTE) (arg & 0xf);
      }

      #ifdef DEBUG
          printk("ioctl(ADC_DIO_PRESET):  value = %#x\n", Chan[minor].dio_value);
      #endif

      break;
    default:
      return(-EINVAL);
      break;
  }
  return 0;
}

/***************************************************************************
 *
 *
 ***************************************************************************/

static struct file_operations adc_fops = {
	NULL,           /* adc_seek */
	adc_read,
	dio_write,      /* dio_write */
	NULL,		/* adc_readdir */
	NULL,		/* adc_select */
	adc_ioctl,	/* adc_ioctl */
	NULL,
	adc_open,
	adc_close,
	NULL		/* fsync */
};

#ifdef __cplusplus
extern "C" {
#endif

/***************************************************************************
 *
 * Loads driver. Called when "insmod adc.o" is invoked on the command line.
 *               The board is set to IRQ, 
 *
 ***************************************************************************/

int init_module(void) 
{
  int i;
  BYTE bReg;

  /* Register as a device with kernel.  */

  if (register_chrdev(MajorNumber, "adc", &adc_fops)) {
    printk("%s: Failure to load module.\n", ADAPTER_ID);
    return -EIO;
  } else {
    printk("%s: address=%#x IRQ=%d.",ADAPTER_ID, BASE_REG, ADC_IRQ);
    printk(" 4/27/95 wjasper@tx.ncsu.edu sam@tx.ncsu.edu\n");
  }
  cli();
  /* Register interrupt handler. */
  if (request_irq(ADC_IRQ, adc0_ReadInterrupt, SA_INTERRUPT, ADAPTER_ID)) {
    sti();
    printk("%s: Can't request IRQ %d\n", ADAPTER_ID, ADC_IRQ);
    return -(EIO);
  }

  /* Set all channel structures to show nothing active/open */

  for (i = 0; i < AD_CHANNELS; i++)
    Chan[i].open = FALSE;
    Chan[i].adc_dio_pretrig = FALSE  ;

  adc0_wait = NULL;

  /* Set interrupt level on board */
#if ADC_IRQ > 7
  bReg = ADC_IRQ-10;
#else
  bReg = ADC_IRQ;
#endif
  bReg <<= 4;                      /* Tell ADC which IRQ to use */
  outb_p(bReg, CNTRL_REG);	   /* ... clearing INTE bit in control reg */

  BoardData.freq = DEFAULT_FREQ;   /* Set default pacer clock frequency */
  SetPacerFreq(&BoardData);
  outb_p(0x0, MISC_REG);           /* Put in compatibility mode */

  sti();

  #ifdef DEBUG
      printk("%s: IRQ set to %d\n", ADAPTER_ID, ADC_IRQ);
  #endif

  return 0;
}

/***************************************************************************
 *
 * Remove driver. Called when "rmmod adc" is run on the command line.
 *
 ***************************************************************************/

void cleanup_module(void) 
{
    if (MOD_IN_USE) {
      printk("%s: device busy, remove delayed\n", ADAPTER_ID);
    }

    free_irq(ADC_IRQ);
    if (unregister_chrdev(MajorNumber, "adc") != 0) {
      printk("%s: cleanup_module failed\n", ADAPTER_ID);
    } else {

    #ifdef DEBUG
        printk("%s: module removed.\n", ADAPTER_ID);
    #endif

    }    
}

#ifdef __cplusplus
}
#endif

/***************************************************************************
 *
 * Block copy of data samples from data register to kernel memory.
 *
 ***************************************************************************/

inline char *RegCopy(char *dest, const int reg, int n)
{
__asm__ __volatile__(
	"cld\n\t"
        "rep\n\t"
        "insw"
	: 
	:"D" (dest), "d" (reg),"c" (n)
	:"di","dx","cx");
return dest;
}

/***************************************************************************
 *
 *  Alarm handler to handle situations where an interrupt is missed.
 *
 ***************************************************************************/

void adc0_TimerHandler(LONG unused)
{
  /* we should only be here if an interrupt occured while we were reading half the FIFO */

  #ifdef DEBUG
      printk("TimerHandler: WordsToRead = %d\n", WordsToRead);
  #endif

  cli();
  outb_p(0x30, COUNTERB_CONTROL);                                 /* Set TOTALBAR to 1 */
  RegCopy((char *) KernBuffPtr, BASE_REG, WordsToRead );
  WordsToRead -= WordsToRead;
  del_timer(&TimerList);
  wake_up_interruptible(&adc0_wait);  
}

/***************************************************************************
 *
 * Interrupt handler used to service enhanced mode(interrupt) read().
 *
 ***************************************************************************/

void adc0_ReadInterrupt(int irq)
{
  BYTE msb, lsb;

  #ifdef DEBUG
      printk("Entering adc0_ReadInterrupt().  mode = %d\n", CurrChan->mode);
  #endif

  cli();
  switch (CurrChan->mode) {

    case ADC_SOFT_TRIGGER:
        /* Bang! Bang! Method: Force conversion, Conversion forces interrupt */
        lsb = inb(BASE_REG);                        /* Sample: ADC -> kernel buffer */
        msb = inb(MSB_DATA_BYTE);
        *KernBuffPtr++ = (msb << 8) | lsb;
        WordsToRead--;                              /* One sample at a time */
        if (!WordsToRead) {
            wake_up_interruptible(&adc0_wait);  
        } else {
            outb_p(0x00, STATUS_REG);	            /* Clear INT bit, allow Interrupt */
            outb_p(0x1, BASE_REG);                  /* Force conversion */
            sti();
        }
        break;

    case ADC_PACER_CLOCK:
    case ADC_EXTERNAL_TRIGGER:
    case ADC_PACER_EXTERN_TRIG:
        if ( WordsToRead > PACKETSIZE ) {
            RegCopy((char *)KernBuffPtr, BASE_REG, PACKETSIZE);
            WordsToRead -= PACKETSIZE;
            KernBuffPtr += PACKETSIZE;
            if ( WordsToRead < PACKETSIZE ) {
                TimerList.expires = WordsToRead*100/BoardData.freq + 2;
                add_timer(&TimerList);
            }
            outb_p(0x00, STATUS_REG);	/* Clear INT bit, allow Interrupt */
            sti();
        } else {
            outb_p(0x30, COUNTERB_CONTROL);         /* Set TOTALBAR to 1 */
            RegCopy((char *) KernBuffPtr, BASE_REG, WordsToRead );
            del_timer(&TimerList);
            WordsToRead -= WordsToRead;
            wake_up_interruptible(&adc0_wait);  
            return;
        }
        break;
    default:
        break;
  }
}

/***************************************************************************
 *
 * Handles software triggered read().
 *
 * Bang! Bang! Method: 
 *    
 *    o Runs in Compatibility Mode
 *    o Force interrupt by forcing conversion.
 *    o Get one sample at a time and put it in the kernel buffer(kernBuff).
 *    o Get chan->count samples
 *
 ***************************************************************************/

static int SoftTrigRead(WORD *kernBuff, ChanRec *chan)
{
  BYTE bReg;

  cli();
  /* Set interrupt level on board, disable interrupts, and pacer  */
#if ADC_IRQ > 7
  bReg = ADC_IRQ-10;
#else
  bReg = ADC_IRQ;
#endif

  bReg <<= 4;                            /* Tell ADC which IRQ to use */
  outb_p(bReg, CNTRL_REG);	         /* & clear INTE bit in control reg */

  bReg = (BYTE) inb_p(MISC_REG);         /* put in compatibility mode */
  bReg &= ~ENHANCED_BIT;
  outb_p(bReg, MISC_REG);

  bReg = (BYTE)inb_p(MUX_SCAN_LIMITS);   /* Read/Write clears FIFO */
  outb_p(bReg, MUX_SCAN_LIMITS);	
  bReg = inb_p(CNTRL_REG);              /* Turns on ADC interrupts */
  outb_p(bReg | CNTL_INTE, CNTRL_REG);
  outb_p(0x00, STATUS_REG);             /* Clears INT bit, allow Int. */

  /* Prepare global data for parameterless interrupt handler */

  CurrChan = chan;                           /* pass chanel number to global */
  KernBuffPtr = KernBuff;                    /* same with KernBuff           */
  WordsToRead = chan->count;

  outb_p(0x1, BASE_REG);                      /* Force first conversion */
  interruptible_sleep_on(&adc0_wait);         /* Block in wait state */

  /* turn off interrupts */

  bReg = inb(CNTRL_REG);
  outb_p(bReg & ~CNTL_INTE, CNTRL_REG);
  return 0;
}

/***************************************************************************
 *
 * Set voltage gains.
 *
 ***************************************************************************/

static void SetGains( BYTE gain )
{
    BYTE mask;

    mask = inb_p(MISC_REG);        /* read the Analog Input Range Register */
    mask &= 0xf0;                  /* mask off last 4 bits */
    mask |= gain;                  /* set gain */
    outb_p(mask, MISC_REG);
}

/****************************************************************************
 *   Set Pacer Clock Frequency
 * 
 *   Description: 
 *       Set the counters so that the pacer clock runs at the
 *       desired frequency.  The frequency is generated by dividing
 *       down a 10 MHz clock, so all frequencies can not be generated.
 *       This routine calculated the divisor to generate a frequency
 *       as near as possible to the requested one.  It then calculates
 *       the real frequency and returns it .
 ****************************************************************************/

static int SetPacerFreq(BoardRec *board)
{
    short unsigned ctr1, ctr2;
    LONG product, error;

    /* divide 10Mhz by frequency */
    product =  10000000.0 / board->freq + 0.5;

    /* Now the job is to find two 16 bit numbers, that when multiplied
       together are approximately equal to product.  Start by setting
       one of them, ctr1 to 2 (minimum settable value) and increment until
       the error is minimized and ctr2 is less than 32768.

       NOTE: In Mode 2, a value of 1 is illegal! Therefore, crt1 and crt2
       can never be 1.

    */

    ctr1 = product / 32768;
    if ( ctr1 < 2 ) ctr1 = 2;
    ctr2 = product / ctr1;
    error = abs(product - ctr2*ctr1);

    while ( error && ctr1 < 32768  && ctr2 > 1 ) {
         ctr1++;
         ctr2 = product / ctr1;
         error = abs(product - ctr2*ctr1);
    }

    /* the frequency is prime, add 1 to it */
    if ( error ) {
        product++;
        ctr1 = product / 32768;
        if ( ctr1 < 2 ) ctr1 = 2;
        ctr2 = product / ctr1;
        error = abs(product - ctr2*ctr1);

        while ( error  && ctr1 < 32768 && ctr2 > 1) {
            ctr1++;
            ctr2 = product / ctr1;
            error = abs(product - ctr2*ctr1);
        }
    }

    /* we can't have ctr2 equal to 1, or system hangs */
    if ( ctr2 == 1 ) {
        ctr2++;
        ctr1 /= 2;
    }

    board->ctr1 = ctr1;
    board->ctr2 = ctr2;
    board->freq = 10000000.0/(ctr1*ctr2) + 0.5;

    #ifdef DEBUG
        printk("SetPacerFreq: Pacer Register set to %#x\n", BoardData.pacerReg);
    #endif

    return 0;
}

/***************************************************************************
 *
 * Load two part frequency to pacer counter chip.
 *
 ***************************************************************************/

static void LoadPacer(BoardRec *board)
{
  BYTE mask;
  BYTE bData;

  /* Write the values of ctr1 and ctr2 into counter A1 and A2 */

  #ifdef DEBUG
      printk("LoadPacer: load values: A ctr1 %#x ctr2 %#x\n", board->ctr1, board->ctr2);
  #endif

  mask = C2+MODE2+LSBFIRST;
  outb_p(mask, COUNTERA_CONTROL); 

  bData = (BYTE) (board->ctr2 & 0xff);
  outb_p(bData, COUNTERA_2_DATA); 

  bData = (BYTE) (board->ctr2 >> 8);
  outb_p(bData, COUNTERA_2_DATA); 

  mask = C1+MODE2+LSBFIRST;
  outb_p(mask, COUNTERA_CONTROL); 

  bData = (board->ctr1 & 0xff);
  outb_p(bData, COUNTERA_1_DATA); 

  bData = (board->ctr1 >> 8) & 0xff;
  outb_p(bData, COUNTERA_1_DATA); 

}

/***************************************************************************
 *
 * Load value into Counter 0  XXXX    Mode    MSB     LSB
 *                            Byte 3  Byte 2  Byte 1  Byte 0
 *
 ***************************************************************************/

static void LoadCounter0( LONG value )
{
  BYTE mask;
  BYTE bData;

  /* Write the value into Counter 0 Mode 2 */

  #ifdef DEBUG
      printk("LoadCounter0: load value %#lx into Counter 0.\n", value);
  #endif

  /* the mode is in the thrid byte */
  mask = 0xff & (value >> 16);
  mask += C0+LSBFIRST;
  outb_p(mask, COUNTERA_CONTROL); 

  if ( value & 0xff000000 ) {       /* load control word only */
      return;

  } else {

      /* LSB in the first byte of value */
      bData = (BYTE) (value & 0xff);
      outb_p(bData, COUNTERA_0_DATA); 

      /* MSB in the second byte of value */
      bData = (BYTE) ((value >> 8) & 0xff);
      outb_p(bData, COUNTERA_0_DATA); 
  }
}

/***************************************************************************
 *
 * Set which channels read() is interested in.
 *
 ***************************************************************************/

static int SetChannelMux(BYTE lowChan, BYTE highChan)
{
    BYTE channelMask;

    channelMask = (highChan << 4) | lowChan;
    outb_p(channelMask, MUX_SCAN_LIMITS);

    return 0;
}

/***************************************************************************
 *
 * Turn off pacer timer chip.
 *
 ***************************************************************************/

static void StopPacer()
{
  BYTE mask;

  mask = C2+MODE2+LSBFIRST;
  outb_p(mask, COUNTERA_CONTROL);
  mask = C1+MODE2+LSBFIRST;
  outb_p(mask, COUNTERA_CONTROL);
}

/***************************************************************************
 *
 * Handles 
 *    o pacer/counter triggered read() with software start.
 *    o pacer/counter triggered read() with external start.
 *    o external triggered read().
 *
 * Pacer Method: 
 *
 *    o Runs in CIO-DAS 16/330 Enhanced Mode
 *    o Pacer counter controls frequency of sample conversions
 *    o Total counter controls number of samples taken
 *    o Interrupts occur every 512(PACKETSIZE) samples and at last sample
 *    o Get chan->count samples
 *    o Samples are saved in kernel buffer(kernBuff) in readInterrupt()
 *
 ***************************************************************************/

static int PacerRead(WORD *kernBuff, ChanRec *chan)
{
  BYTE bReg = 0x0;
  WORD wReg = 0x0;

  #ifdef DEBUG
      printk("Entering PacerRead().\n");
  #endif

  cli();

  /* Set interrupt level on board, disable interrupts, and pacer  */
#if ADC_IRQ > 7
  bReg = ADC_IRQ-10;
#else
  bReg = ADC_IRQ;
#endif

  bReg <<= 4;                                /* Tell ADC which IRQ to use */
  outb_p(bReg, CNTRL_REG);	             /* & clear INTE bit in control reg */

  bReg = (BYTE)inb_p(MUX_SCAN_LIMITS);       /* Read/Write clears FIFO */
  outb_p(bReg, MUX_SCAN_LIMITS);	

  bReg = chan->gain | ENHANCED_BIT;          /* Set gain and enhanced mode */
  outb(bReg, MISC_REG);
  bReg = (BYTE)inb_p(MISC_REG);	             /* check to make sure we are in */
  if (!(bReg & ENHANCED_BIT)) {	             /* enhanced mode                */
    return(NO_ENHANCED_MODE);
  }

  bReg = (BYTE)inb_p(MUX_SCAN_LIMITS);       /* Read/Write clears FIFO */
  outb_p(bReg, MUX_SCAN_LIMITS);	

  if (DT_CONNECT) {
    bReg = (BYTE) inb_p(MISC_REG);    
    bReg |= MISC_DT;
    outb_p(bReg, MISC_REG);
  }

  /* load the Total Counter  for the DAS 16/330 */

  bReg = C0+MODE0+LSBFIRST;	             /* CounterB -> mode 0, DON'T load it */
  outb_p(bReg, COUNTERB_CONTROL);            /* Sets output HIGH? */

  bReg = (BYTE)inb_p(CNTRL_REG);             /* S1=0 and S0=0 from above */
  if( chan->mode == ADC_PACER_CLOCK || chan->mode == ADC_PACER_EXTERN_TRIG ) {
      bReg |= (0x3);                         /* Set S1=1 and S0=1 Internal Pacer */
  } else {
      bReg |= (0x2);                         /* Set S1=1 and S0=0 External Pacer (A/D Triggering) */
  }
  outb_p(bReg, CNTRL_REG);

  wReg = (WORD) (chan->count >> 16);         /* get upper word of count */
  if (chan->count & 0xffff) wReg++;          /* increment if lower count is ! zero */

  bReg = C0+MODE0+LSBFIRST;	             /* Counter -> mode 0, DON'T load it */
  outb_p(bReg, COUNTERB_CONTROL);            /* Sets output HIGH? */
  bReg = (BYTE) (wReg & 0xff);	             /* load the upper word */
  outb_p(bReg, COUNTERB_0_DATA);
  bReg = (BYTE) (wReg >> 8);
  outb_p(bReg, COUNTERB_0_DATA);

  /* Set the trigger counters. Counter 0 is changed to counter 1 */

  bReg = C1+MODE2+LSBFIRST;	             /* Ensure clock starts low */
  outb_p(bReg, COUNTERB_CONTROL);    

  bReg = C1+MODE0+LSBFIRST;	             /* Make C0 clock high */
  outb_p(bReg, COUNTERB_CONTROL); 

  bReg = C1+MODE2+LSBFIRST;	             /* Make C0 clock low */
  outb_p(bReg, COUNTERB_CONTROL);            /* complete 1st clock cycle */

  wReg =  (WORD) (chan->count & 0xffff);

  if (wReg == 1 ) {
    wReg++;			             /* wReg can not be 1 */
    bReg = (BYTE) (wReg & 0xff);             /* load the lower word */
    outb_p(bReg, COUNTERB_1_DATA);
    bReg = (BYTE) (wReg >> 8);
    outb_p(bReg, COUNTERB_1_DATA);
    outb_p(0x1, BASE_REG);	             /* Force conversion to get count = 1 */
  } else {
    bReg = (BYTE) (wReg & 0xff);             /* load the lower word */
    outb_p(bReg, COUNTERB_1_DATA);
    bReg = (BYTE) (wReg >> 8);
    outb_p(bReg, COUNTERB_1_DATA);
  }
  outb_p(0x1, BASE_REG);	             /* Force a conversion */
  outw_p(0x0, COUNTERB_1_DATA);              /* rollover mode ?? */

  if( chan->mode == ADC_PACER_CLOCK || chan->mode == ADC_PACER_EXTERN_TRIG ) {
      LoadPacer(&BoardData);                 /* Establish sample frequency */
  }

  bReg = (BYTE)inb_p(MUX_SCAN_LIMITS);       /* Read/Write clears FIFO */
  outb_p(bReg, MUX_SCAN_LIMITS);	

  outb_p(0x00, STATUS_REG);	             /* Clear INT bit, allow Interrupt */

  /* Prepare global data for parameterless interrupt handler */

  CurrChan = chan;                              /* pass chanel number to global */
  KernBuffPtr = KernBuff;                       /* same with KernBuff           */
  WordsToRead = chan->count;

  bReg = (BYTE)inb_p(CNTRL_REG);                /* Enable interrupts */
  bReg |= CNTL_INTE;           
  outb_p(bReg, CNTRL_REG);

  if( chan->mode == ADC_PACER_CLOCK || chan->mode == ADC_PACER_EXTERN_TRIG ) {
      outb_p(BoardData.pacerReg, PACER_CLOCK_REG);  /* Start pacer begins sampling. */
  }

  if ( chan->adc_dio_pretrig ) {
      outb_p( chan->dio_value, DIO_REG );
  }

  interruptible_sleep_on(&adc0_wait);           /* Block in wait state */

  /* turn off interrupts, and trigger control.  Leave pacer clock on if enabled */
  outb_p(0x0, CNTRL_REG);

  if ( WordsToRead != 0 ) {
      printk("Timing error in PacerRead: WordsToRead = %d\n", WordsToRead);
      return -1;
  }
  outb_p(0x00, STATUS_REG);             /* Clear INT bit */
  return 0;
}

